import glob
import json
import logging
import os
import sys
from collections import OrderedDict
from typing import Union, List, Dict, Tuple

import numpy as np
from smac.epm.rfr_imputator import RFRImputator
from smac.epm.util_funcs import get_types
from smac.tae import StatusType
from smac.utils.io.traj_logging import TrajLogger
from tqdm import tqdm

from pimp.configspace import CategoricalHyperparameter, Configuration, impute_inactive_values
from pimp.epm.base_epm import RandomForestWithInstances
from pimp.epm.unlogged_epar_x_rfwi import UnloggedEPARXrfi
from pimp.epm.unlogged_rfwi import Unloggedrfwi
from pimp.evaluator.ablation import Ablation
from pimp.evaluator.fanova import fANOVA
from pimp.evaluator.forward_selection import ForwardSelector, AbstractEvaluator
from pimp.evaluator.influence_models import InfluenceModel
from pimp.evaluator.local_parameter_importance import LPI
from pimp.utils import RunHistory, RunHistory2EPM4Cost, RunHistory2EPM4LogCost, Scenario

__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


class Importance(object):
    def __init__(self, scenario_file: Union[None, str] = None, scenario: Union[None, Scenario] = None,
                 runhistory_file: Union[str, None] = None, runhistory: Union[None, RunHistory] = None,
                 traj_file: Union[None, List[str]] = None, incumbent: Union[None, Configuration] = None,
                 seed: int = 12345, parameters_to_evaluate: int = -1, margin: Union[None, float] = None,
                 save_folder: str = 'PIMP', impute_censored: bool = False, max_sample_size: int = -1,
                 fANOVA_cut_at_default=False, fANOVA_pairwise=True, forwardsel_feat_imp=False,
                 incn_quant_var=True, preprocess=False, forwardsel_cv=False, verbose: bool=True):
        """
        Importance Object. Handles the construction of the data and training of the model. Easy interface to the
        different evaluators.
        :param scenario_file: File to load the scenario from, if scenario is None.
        :param scenario: Scenario Object to use if scenario_file is None
        :param runhistory_file: File to load the runhistory from if runhistory is None.
        :param runhistory: Runhistory Object to use if runhistory_file is None.
        :param traj_file: File to load the trajectory from. If this is None but runhistory_file was specified,
               the trajectory will be read from the same directory as the runhistory. (Will be ignored if incumbent is
               set)
        :param incumbent: Configuration Object to use.
        :param seed: Seed used for the numpy random generator.
        :param parameters_to_evaluate: int that specifies how many parameters have to be evaluated.
               If set to -1 all parameters will be evaluated.
        :param margin: float used in conjunction with influence models. Is the minimal improvement to accept a
                       parameter as important.
        :param save_folder: Folder name to save the output to
        :param impute_censored: boolean that specifies if censored data should be imputed. If not, censored data are
               ignored.
        :param verbose: Toggle output to stdout (not logging, but tqdm-progress bars)
        """
        self.logger = logging.getLogger("pimp.Importance")
        self.rng = np.random.RandomState(seed)
        self._parameters_to_evaluate = parameters_to_evaluate
        self._evaluator = None
        self.margin = margin
        self.threshold = None
        self.seed = seed
        self.impute = impute_censored
        self.cut_def_fan = fANOVA_cut_at_default
        self.pairiwse_fANOVA = fANOVA_pairwise
        self.forwardsel_feat_imp = forwardsel_feat_imp
        self.incn_quant_var = incn_quant_var
        self.preprocess = preprocess
        self._preprocessed = False
        self.X_fanova = None
        self.y_fanova = None
        self.forwardsel_cv = forwardsel_cv
        self.verbose = verbose

        self.evaluators = []

        self._setup_scenario(scenario, scenario_file, save_folder)
        self._load_runhist(runhistory, runhistory_file)
        self._setup_model()
        self.best_dir = None
        self._load_incumbent(traj_file, runhistory_file, incumbent)
        self.logger.info('Best incumbent found in %s' % self.best_dir)
        if 0 < max_sample_size < len(self.X):
            self.logger.warning('Reducing the amount of datapoints!')
            if self.best_dir:
                self.logger.warning('Only using the runhistory that contains the incumbent!')
                self.logger.info('Loading from %s' % self.best_dir)
                self._load_runhist(None, os.path.join(self.best_dir, '*history*'))
                self._convert_data(fit=False)
                self._load_incumbent(glob.glob(os.path.join(self.best_dir, '*traj_aclib2*'), recursive=True)[0], None,
                                     incumbent)
                if max_sample_size < len(self.X):
                    self.logger.warning('Also downsampling as requested!')
                    idx = list(range(len(self.X)))
                    np.random.shuffle(idx)
                    idx = idx[:max_sample_size]
                    self.X = self.X[idx]
                    self.y = self.y[idx]
            else:
                self.logger.warning('Downsampling as requested!')
                idx = list(range(len(self.X)))
                np.random.shuffle(idx)
                idx = idx[:max_sample_size]
                self.X = self.X[idx]
                self.y = self.y[idx]
            self.logger.info('Remaining %d datapoints' % len(self.X))
            self.model.train(self.X, self.y)

    def _preprocess(self, runhistory):
        """
        Method to marginalize over instances such that fANOVA can determine the parameter importance without
        having to deal with instance features.
        :param runhistory: RunHistory that knows all configurations that were run. For all these configurations
                           we have to marginalize away the instance features with which fANOVA will make it's
                           predictions
        """
        self.logger.info('PREPROCESSING PREPROCESSING PREPROCESSING PREPROCESSING PREPROCESSING PREPROCESSING')
        self.logger.info('Marginalizing away all instances!')
        configs = runhistory.get_all_configs()
        X_non_hyper, X_prime, y_prime = [], [], []
        for c_id, config in tqdm(enumerate(configs), ascii=True, desc='Completed: ', total=len(configs)):
            config = impute_inactive_values(config).get_array()
            X_prime.append(config)
            X_non_hyper.append(config)
            y_prime.append(self.model.predict_marginalized_over_instances(np.array([X_prime[-1]]))[0].flatten())
            for idx, param in enumerate(self.scenario.cs.get_hyperparameters()):
                if not isinstance(param, CategoricalHyperparameter):
                    X_non_hyper[-1][idx] = param._transform(X_non_hyper[-1][idx])
        X_non_hyper = np.array(X_non_hyper)
        X_prime = np.array(X_prime)
        y_prime = np.array(y_prime)
        # y_prime = np.array(self.model.predict_marginalized_over_instances(X_prime)[0])
        self.X = X_prime
        self.X_fanova = X_non_hyper
        self.y_fanova = y_prime
        self.y = y_prime
        self.logger.info('Size of training X after preprocessing: %s' % str(self.X.shape))
        self.logger.info('Size of training y after preprocessing: %s' % str(self.y.shape))
        self.logger.info('Finished Preprocessing')
        self._preprocessed = True

    def _setup_scenario(self, scenario: Union[None, Scenario], scenario_file: Union[None, str], save_folder: str) -> \
            None:
        """
        Setup for the scenario
        Helper method to have the init method less cluttered.
        For parameter specifications, see __init__
        """
        if scenario is not None:
            self.scenario = scenario
        elif scenario_file is not None:
            self.logger.info('Reading Scenario file and files specified in the scenario')
            self.scenario = Scenario(scenario=scenario_file)
            self.scenario.output_dir = save_folder
            self.scenario.output_dir_for_this_run = save_folder
            written = self.scenario.out_writer.write_scenario_file(self.scenario)
        else:
            raise Exception('Either a scenario has to be given or a file to load it from! Both were set to None!')

    def _load_incumbent(self, traj_file, runhistory_file, incumbent, predict_best=True) -> None:
        """
        Handles the loading of the incumbent according to the given parameters.
        Helper method to have the init method less cluttered.
        For parameter specifications, see __init__
        """
        self.incumbent = (None, None)
        if incumbent is not None:
            self.incumbent = incumbent
        elif traj_file is not None:
            self.logger.info('Reading traj_file: %s' % traj_file)
            self.incumbent = self._read_traj_file(traj_file)[0]
            self.logger.debug('Incumbent %s' % str(self.incumbent))
        elif traj_file is None and runhistory_file is not None:
            traj_files = os.path.join(os.path.dirname(runhistory_file), 'traj_aclib2.json')
            traj_files = sorted(glob.glob(traj_files, recursive=True))
            incumbents = []
            for traj_ in traj_files:
                self.logger.info('Reading traj_file: %s' % traj_)
                incumbents.append(self._read_traj_file(traj_))
                incumbents[-1].extend(self._model.predict_marginalized_over_instances(
                    np.array([impute_inactive_values(incumbents[-1][0]).get_array()])))
                self.logger.debug(incumbents[-1])
            sort_idx = 2 if predict_best else 1
            incumbents = sorted(enumerate(incumbents), key=lambda x: x[1][sort_idx])
            self.best_dir = os.path.dirname(traj_files[incumbents[0][0]])
            self.incumbent = incumbents[0][1][0]
            self.logger.info('Incumbent %s' % str(self.incumbent))
        else:
            raise Exception('No method specified to load an incumbent. Either give the incumbent directly or specify '
                            'a file to load it from!')

    def _setup_model(self) -> None:
        """
        Sets up all the necessary parameters used for the model.
        Helper method to have the init method less cluttered.
        For parameter specifications, see __init__
        """
        self.logger.info('Converting Data and constructing Model')
        self.X = None
        self.y = None
        self.types = None
        self.bounds = None
        self._model = None
        self.logged_y = False
        self._convert_data(fit=True)
        if self.preprocess:
            self._preprocess(self.runhistory)
            if self.scenario.run_obj == "runtime":
                self.y = np.log10(self.y)
            self.model = 'urfi'
            self.model.train(self.X, self.y)

    def _load_runhist(self, runhistory, runhistory_file) -> None:
        """
        Handels loading of the runhistory/runhistories.
        Helper method to have the init method less cluttered.
        For parameter specifications, see __init__
        """
        self.logger.debug(runhistory_file)
        self.logger.debug(runhistory)
        if runhistory is not None:
            self.runhistory = runhistory
        elif runhistory_file is not None:
            self.logger.info('Reading Runhistory')
            self.runhistory = RunHistory()

            globed_files = glob.glob(runhistory_file)
            self.logger.info('#RunHistories found: %d' % len(globed_files))
            if not globed_files:
                self.logger.error('No runhistory files found!')
                sys.exit(1)
            self.runhistory.load_json(globed_files[0], self.scenario.cs)
            for rh_file in globed_files[1:]:
                self.runhistory.update_from_json(rh_file, self.scenario.cs)
        else:
            raise Exception('Either a runhistory or files to load them from have to be specified! Both were set to '
                            'None!')
        self.logger.info('Combined number of Runhistory data points: %d' % len(self.runhistory.data))
        self.logger.info('Number of Configurations: %d' % (len(self.runhistory.get_all_configs())))

    def _read_traj_file(self, fn):
        """
        Simple method to read in a trajectory file in the json format / aclib2 format
        :param fn:
            file name
        :return:
            tuple of (incumbent [Configuration], incumbent_cost [float])
        """
        if not (os.path.exists(fn) and os.path.isfile(fn)):  # File existence check
            raise FileNotFoundError('File %s not found!' % fn)
        with open(fn) as fp:
            # In aclib2, the incumbent is a list of strings, in alljson it's a dictionary.
            fileformat = 'aclib2' if isinstance(json.loads(fp.readline())["incumbent"], list) else 'alljson'

        if fileformat == "aclib2":
            self.logger.info("Format is 'aclib2'. This format has issues with recovering configurations properly. We "
                             "recommend to use the alljson-format.")
            traj = TrajLogger.read_traj_aclib_format(fn, self.scenario.cs)
        else:
            traj = TrajLogger.read_traj_alljson_format(fn, self.scenario.cs)

        incumbent_cost = traj[-1]['cost']
        incumbent = traj[-1]['incumbent']
        return [incumbent, incumbent_cost]

    @property
    def model(self):
        return self._model

    def _get_types(self, scenario, features):
        types, bounds = get_types(scenario, features)
        types = np.array(types, dtype='uint')
        bounds = np.array(bounds, dtype='object')
        return types, bounds

    @model.setter
    def model(self, model_short_name='urfi'):
        if model_short_name not in ['urfi', 'rfi']:
            raise ValueError('Specified model %s does not exist or not supported!' % model_short_name)
        elif model_short_name == 'rfi':
            self.types, self.bounds = self._get_types(self.scenario.cs, self.scenario.feature_array)
            self._model = RandomForestWithInstances(self.scenario.cs, self.types, self.bounds, 12345,
                                                    instance_features=self.scenario.feature_array,
                                                    logged_y=self.logged_y)
        elif model_short_name == 'urfi':
            self.logged_y = True
            if not self._preprocessed:
                self.types, self.bounds = self._get_types(self.scenario.cs, self.scenario.feature_array)
                self._model = UnloggedEPARXrfi(self.scenario.cs, self.types, self.bounds, 12345,
                                               instance_features=self.scenario.feature_array,
                                               cutoff=self.cutoff, threshold=self.threshold,
                                               logged_y=self.logged_y)
            else:
                self.types, self.bounds = self._get_types(self.scenario.cs, None)
                self._model = Unloggedrfwi(self.scenario.cs, self.types, self.bounds, 12345,
                                           instance_features=None,
                                           logged_y=self.logged_y)
        self._model.rf_opts.compute_oob_error = True

    @property
    def evaluator(self) -> AbstractEvaluator:
        """
        Getter of the evaluator property. Returns the set evaluation method.
        :return: AbstractEvaluator
        """
        return self._evaluator

    @evaluator.setter
    def evaluator(self, evaluation_method: str) -> None:
        """
        Setter of the evaluator property. The wanted evaluation method can be specified as string and the rest is
        handled automatically here
        :param evaluation_method: Name of the evaluation method to use
        :return: None
        """
        if self._model is None:
            self._setup_model()
        self.logger.info('Setting up Evaluation Method')
        if evaluation_method not in ['ablation', 'fanova', 'forward-selection', 'influence-model',
                                     'incneighbor', 'lpi']:
            raise ValueError('Specified evaluation method %s does not exist!' % evaluation_method)
        if evaluation_method == 'ablation':
            if self.incumbent is None:
                raise ValueError('Incumbent is %s!\n \
                                 Incumbent has to be read from a trajectory file before ablation can be used!'
                                 % self.incumbent)
            self.logger.info('Using model %s' % str(self.model))
            self.logger.info('X shape %s' % str(self.model.X.shape))
            evaluator = Ablation(scenario=self.scenario,
                                 cs=self.scenario.cs,
                                 model=self._model,
                                 to_evaluate=self._parameters_to_evaluate,
                                 incumbent=self.incumbent,
                                 logy=self.logged_y,
                                 rng=self.rng,
                                 verbose=self.verbose)
        elif evaluation_method == 'influence-model':
            self.logger.info('Using model %s' % str(self.model))
            self.logger.info('X shape %s' % str(self.model.X.shape))
            evaluator = InfluenceModel(scenario=self.scenario,
                                       cs=self.scenario.cs,
                                       model=self._model,
                                       to_evaluate=self._parameters_to_evaluate,
                                       margin=self.margin,
                                       threshold=self.threshold,
                                       rng=self.rng,
                                       verbose=self.verbose)
        elif evaluation_method == 'fanova':
            self.logger.info('Using model %s' % str(self.model))
            self.logger.info('X shape %s' % str(self.model.X.shape))
            mini = None
            if self.cut_def_fan:
                mini = True         # TODO what about scenarios where we maximize?
            evaluator = fANOVA(scenario=self.scenario,
                               cs=self.scenario.cs,
                               model=self._model,
                               to_evaluate=self._parameters_to_evaluate,
                               runhist=self.runhistory,
                               rng=self.rng,
                               minimize=mini,
                               pairwise=self.pairiwse_fANOVA,
                               preprocessed_X=self.X_fanova,
                               preprocessed_y=self.y_fanova,
                               incumbents=self.incumbent,
                               verbose=self.verbose)
        elif evaluation_method in ['incneighbor', 'lpi']:
            if self.incumbent is None:
                raise ValueError('Incumbent is %s!\n \
                                 Incumbent has to be read from a trajectory file before LPI can be used!'
                                 % self.incumbent)
            self.logger.info('Using model %s' % str(self.model))
            self.logger.info('X shape %s' % str(self.model.X.shape))
            evaluator = LPI(scenario=self.scenario,
                            cs=self.scenario.cs,
                            model=self._model,
                            to_evaluate=self._parameters_to_evaluate,
                            incumbent=self.incumbent,
                            logy=self.logged_y,
                            rng=self.rng,
                            quant_var=self.incn_quant_var,
                            verbose=self.verbose)
        else:
            self.logger.info('Using model %s' % str(self.model))
            evaluator = ForwardSelector(scenario=self.scenario,
                                        cs=self.scenario.cs,
                                        model=self._model,
                                        to_evaluate=self._parameters_to_evaluate,
                                        rng=self.rng,
                                        feature_imp=self.forwardsel_feat_imp,
                                        cv=self.forwardsel_cv,
                                        verbose=self.verbose)
        self._evaluator = evaluator

    def _convert_data(self, fit=True) -> None:  # From Marius
        '''
            converts data from runhistory into EPM format

            Parameters
            ----------
            scenario: Scenario
                smac.scenario.scenario.Scenario Object
            runhistory: RunHistory
                smac.runhistory.runhistory.RunHistory Object with all necessary data

            Returns
            -------
            np.array
                X matrix with configuartion x features for all observed samples
            np.array
                y matrix with all observations
            np.array
                types of X cols -- necessary to train our RF implementation
        '''

        params = self.scenario.cs.get_hyperparameters()
        num_params = len(params)
        self.logger.debug("Counted %d hyperparameters", num_params)

        if self.scenario.run_obj == "runtime":
            self.cutoff = self.scenario.cutoff
            self.threshold = self.scenario.cutoff * self.scenario.par_factor
            self.model = 'urfi'
            self.logged_y = True
            # if we log the performance data,
            # the RFRImputator will already get
            # log transform data from the runhistory
            cutoff = np.log10(self.scenario.cutoff)
            threshold = np.log10(self.scenario.cutoff *
                                 self.scenario.par_factor)
            model = RandomForestWithInstances(self.scenario.cs,
                                              self.types, self.bounds, 12345,
                                              instance_features=self.scenario.feature_array,
                                              )

            imputor = RFRImputator(rng=self.rng,
                                   cutoff=cutoff,
                                   threshold=threshold,
                                   model=model,
                                   change_threshold=0.01,
                                   max_iter=10)
            rh2EPM = RunHistory2EPM4LogCost(scenario=self.scenario,
                                            num_params=num_params,
                                            success_states=[
                                                StatusType.SUCCESS, ],
                                            impute_censored_data=self.impute,
                                            impute_state=[
                                                StatusType.TIMEOUT, StatusType.CAPPED],
                                            imputor=imputor)
        else:
            self.model = 'rfi'
            rh2EPM = RunHistory2EPM4Cost(scenario=self.scenario,
                                         num_params=num_params,
                                         success_states=[StatusType.SUCCESS],
                                         impute_censored_data=self.impute,
                                         impute_state=None)
        self.logger.info('Using model %s' % str(self.model))
        X, Y = rh2EPM.transform(self.runhistory)

        self.X = X
        self.y = Y
        self.logger.info('Size of training X: %s' % str(self.X.shape))
        self.logger.info('Size of training y: %s' % str(self.y.shape))
        self.logger.info('Data was %s imputed' % ('not' if not self.impute else ''))
        if not self.impute:
            self.logger.info('Thus the size of X might be smaller than the datapoints in the RunHistory')
        if fit:
            self.logger.info('Fitting Model')
            self.model.train(X, Y)

    def evaluate_scenario(self, methods, save_folder=None, plot_pyplot=True, plot_bokeh=False) -> Union[
            Tuple[Dict[str, Dict[str, float]], List[AbstractEvaluator]], Dict[str, Dict[str, float]]]:
        """
         the given scenario
        :param evaluation_method: name of the method to use
        :param sort_by: int, determines the order (only used if evaluation_method == all)
            0 => Ablation, fANOVA, Forward Selection
            1 => Ablation, Forward Selection, fANOVA
            2 => fANOVA, Forward Selection, Ablation
            3 => fANOVA, Ablation, Forward Selection
            4 => Forward Selection, Ablation, fANOVA
            5 => Forward Selection, fANOVA, Ablation
        :param plot_pyplot: whether to perform standard matplotlib- plotting
        :param plot_bokeh: whether to perform advanced bokeh plotting
        :return: if evaluation all: Tupel of dictionary[evaluation_method] -> importance values, List ov evaluator
                                    names, ordered according to sort_by
                 else:
                      dict[evalution_method] -> importance values
        """
        # influence-model currently not supported
        if not len(methods) >= 1:
            raise ValueError("Specify at least one method to evaluate the scenario!")
        fn = os.path.join(save_folder, 'pimp_results.json')
        load = os.path.exists(fn)
        dict_ = {}
        for rnd, method in enumerate(methods):
            self.logger.info('Running %s' % method)
            self.evaluator = method
            dict_[self.evaluator.name.lower()] = self.evaluator.run()
            self.evaluators.append(self.evaluator)
            if save_folder and plot_pyplot:
                self.evaluator.plot_result(os.path.join(save_folder, self.evaluator.name.lower()), show=False)
            if save_folder and plot_bokeh:
                self.evaluator.plot_bokeh(os.path.join(save_folder, self.evaluator.name.lower() + "_bokeh"))
            if load:
                with open(fn, 'r') as in_file:
                    doct = json.load(in_file)
                    for key in doct:
                        dict_[key] = doct[key]
            if save_folder:
                with open(fn, 'w') as out_file:
                    json.dump(dict_, out_file, sort_keys=True, indent=4, separators=(',', ': '))
                    load = True
        return dict_, self.evaluators

    def plot_results(self, name: Union[List[str], str, None] = None, evaluators: Union[List[AbstractEvaluator],
                                                                                       None] = None,
                     show: bool = True):
        """
        Method to handle the plotting in case of plots for multiple evaluation methods or only one
        :param name: name(s) to save the plot(s) with
        :param evaluators: list of ealuators to generate the plots for
        :param show: boolean. Specifies if the results have to additionally be shown and not just saved!
        :return:
        """
        if evaluators:
            for eval, name_ in zip(evaluators, name):
                eval.plot_result(name_, show)
        else:
            self.evaluator.plot_result(name, show)

    def table_for_comparison(self, evaluators: List[AbstractEvaluator], name: Union[None, str] = None, style='cmd'):
        """
        Small Method that creates an output table for comparison either printed in a readable format for the command
        line or in latex style
        :param evaluators: All evaluators to put into the table
        :param name: Name for the save file name
        :param style: (cmd|latex) str to determine which format to use
        :return: None
        """
        if name:
            f = open(name, 'w')
        else:
            f = sys.stderr
        header = ['{:>{width}s}' for _ in range(len(evaluators) + 1)]
        line = '-' if style == 'cmd' else '\hline'
        join_ = ' | ' if style == 'cmd' else ' & '
        body = OrderedDict()
        _max_len_p = 1
        _max_len_h = 1
        for idx, e in enumerate(evaluators):
            for p in e.evaluated_parameter_importance:
                if p not in ['-source-', '-target-']:
                    if p not in body:
                        body[p] = ['-' for _ in range(len(evaluators))]
                        body[p][idx] = e.evaluated_parameter_importance[p]
                        _max_len_p = max(_max_len_p, len(p))
                    else:
                        body[p][idx] = e.evaluated_parameter_importance[p]
                    if e.name in ['Ablation', 'fANOVA', 'LPI']:
                        if body[p][idx] != '-':
                            body[p][idx] *= 100
                        _max_len_p = max(_max_len_p, len(p))

            header[idx + 1] = e.name
            _max_len_h = max(_max_len_h, len(e.name))
        header[0] = header[0].format(' ', width=_max_len_p)
        header[1:] = list(map(lambda x: '{:^{width}s}'.format(x, width=_max_len_h), header[1:]))
        header = join_.join(header)
        if style == 'latex':
            print('\\begin{table}', file=f)
            print('\\begin{tabular}{r%s}' % ('|r' * len(evaluators)), file=f)
            print('\\toprule', file=f)
        print(header, end='\n' if style == 'cmd' else '\\\\\n', file=f)
        if style == 'cmd':
            print(line * len(header), file=f)
        else:
            print(line, file=f)
        for p in body:
            if style == 'cmd':
                b = ['{:>{width}s}'.format(p, width=_max_len_p)]
            else:
                b = ['{:<{width}s}'.format(p, width=_max_len_p)]
            for x in body[p]:
                try:
                    if style == 'latex':
                        b.append('${:> {width}.3f}$'.format(x, width=_max_len_h - 2))
                    else:
                        b.append('{:> {width}.3f}'.format(x, width=_max_len_h))
                except ValueError:
                    b.append('{:>{width}s}'.format(x, width=_max_len_h))
            print(join_.join(b), end='\n' if style == 'cmd' else '\\\\\n', file=f)
        cap = 'Parameter Importance values, obtained using the PIMP package. Ablation values are percentages ' \
              'of improvement a single parameter change obtained between the default and an' \
              ' incumbent configuration.\n' \
              'fANOVA values are percentages that show how much variance across the whole ConfigSpace can be ' \
              'explained by that parameter.\n' \
              'Forward Selection values are RMSE values obtained using only a subset of parameters for prediction.\n' \
              'fANOVA and Forward Selection try to estimate the importances across the whole parameter space, while ' \
              'ablation tries to estimate them between two given configurations.'
        if self._parameters_to_evaluate > 0:
            cap += """\nOnly the top %d parameters of each method are listed.
                    "-" represent that this parameter was not evaluated
                     using the given method but with another.
                    """ % self._parameters_to_evaluate
        if style == 'latex':
            print('\\bottomrule', file=f)
            print('\end{tabular}', file=f)
            print('\\caption{%s}' % cap,
                  file=f)
            print('\\label{tab:pimp}', file=f)
            print('\end{table}', file=f)
        else:
            print('', file=f)
            print(cap)
        if name:
            f.close()



import numpy as np
import logging

from pyrfr import regression

from smac.epm.base_epm import AbstractEPM
from smac.configspace import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
    Constant,
)


class RandomForestWithInstances(AbstractEPM):
    """
    Base EPM. Very similar to SMAC3s EPM.
    Interface to the random forest that takes instance features
    into account.

    Attributes
    ----------
    rf_opts :
        Random forest hyperparameter
    n_points_per_tree : int
    rf : regression.binary_rss_forest
        Only available after training
    hypers: list
        List of random forest hyperparameters
    seed : int
    types : list
    bounds : list
    rng : np.random.RandomState
    logger : logging.logger
    """

    def __init__(self,
                 configspace,
                 types: np.ndarray,
                 bounds: np.ndarray,
                 seed: int,
                 num_trees: int = 10,
                 do_bootstrapping: bool = True,
                 n_points_per_tree: int = -1,
                 ratio_features: float = 5. / 6.,
                 min_samples_split: int = 3,
                 min_samples_leaf: int = 3,
                 max_depth: int = 20,
                 eps_purity: int = 1e-8,
                 max_num_nodes: int = 2 ** 20,
                 logged_y: bool = True,
                 **kwargs):
        """Constructor

        Parameters
        ----------
        configspace: ConfigurationSpace
            configspace to be passed to random forest (used to impute inactive parameter-values)
        types : np.ndarray (D)
            Specifies the number of categorical values of an input dimension where
            the i-th entry corresponds to the i-th input dimension. Let's say we
            have 2 dimension where the first dimension consists of 3 different
            categorical choices and the second dimension is continuous than we
            have to pass np.array([2, 0]). Note that we count starting from 0.
        bounds : np.ndarray (D, 2)
            Specifies the bounds for continuous features.
        seed : int
            The seed that is passed to the random_forest_run library.
        num_trees : int
            The number of trees in the random forest.
        do_bootstrapping : bool
            Turns on / off bootstrapping in the random forest.
        n_points_per_tree : int
            Number of points per tree. If <= 0 X.shape[0] will be used
            in _train(X, y) instead
        ratio_features : float
            The ratio of features that are considered for splitting.
        min_samples_split : int
            The minimum number of data points to perform a split.
        min_samples_leaf : int
            The minimum number of data points in a leaf.
        max_depth : int
            The maximum depth of a single tree.
        eps_purity : float
            The minimum difference between two target values to be considered
            different
        max_num_nodes : int
            The maxmimum total number of nodes in a tree
        logged_y: bool
            Indicates if the y data is transformed (i.e. put on logscale) or not
        """
        super().__init__(configspace=configspace, types=types, bounds=bounds, seed=seed, **kwargs)


        self.configspace = configspace
        self.types = types
        self.bounds = bounds
        self.rng = regression.default_random_engine(seed)

        self.rf_opts = regression.forest_opts()
        self.rf_opts.num_trees = num_trees
        self.rf_opts.do_bootstrapping = do_bootstrapping
        max_features = 0 if ratio_features > 1.0 else \
            max(1, int(types.shape[0] * ratio_features))
        self.rf_opts.tree_opts.max_features = max_features
        self.rf_opts.tree_opts.min_samples_to_split = min_samples_split
        self.rf_opts.tree_opts.min_samples_in_leaf = min_samples_leaf
        self.rf_opts.tree_opts.max_depth = max_depth
        self.rf_opts.tree_opts.epsilon_purity = eps_purity
        self.rf_opts.tree_opts.max_num_nodes = max_num_nodes
        self.rf_opts.compute_law_of_total_variance = False  # Always off. No need for this in our base EPM

        self.n_points_per_tree = n_points_per_tree
        self.rf = None  # type: regression.binary_rss_forest
        self.logged_y = logged_y

        # This list well be read out by save_iteration() in the solver
        self.hypers = [num_trees, max_num_nodes, do_bootstrapping,
                       n_points_per_tree, ratio_features, min_samples_split,
                       min_samples_leaf, max_depth, eps_purity, seed]
        self.seed = seed

        self.impute_values = {}

        self.logger = logging.getLogger(self.__module__ + "." +
                                        self.__class__.__name__)


    def _impute_inactive(self, X: np.ndarray) -> np.ndarray:
        X = X.copy()
        for idx, hp in enumerate(self.configspace.get_hyperparameters()):
            if idx not in self.impute_values:
                parents = self.configspace.get_parents_of(hp.name)
                if len(parents) == 0:
                    self.impute_values[idx] = None
                else:
                    if isinstance(hp, CategoricalHyperparameter):
                        self.impute_values[idx] = len(hp.choices)
                    elif isinstance(hp, (UniformFloatHyperparameter, UniformIntegerHyperparameter)):
                        self.impute_values[idx] = -1
                    elif isinstance(hp, Constant):
                        self.impute_values[idx] = 1
                    else:
                        raise ValueError

            nonfinite_mask = ~np.isfinite(X[:, idx])
            X[nonfinite_mask, idx] = self.impute_values[idx]

        return X

    def _train(self, X: np.ndarray, y: np.ndarray, **kwargs):
        """Trains the random forest on X and y.

        Parameters
        ----------
        X : np.ndarray [n_samples, n_features (config + instance features)]
            Input data points.
        Y : np.ndarray [n_samples, ]
            The corresponding target values.

        Returns
        -------
        self
        """

        self.X = self._impute_inactive(X)
        self.y = y.flatten()
        if self.n_points_per_tree <= 0:
            self.rf_opts.num_data_points_per_tree = self.X.shape[0]
        else:
            self.rf_opts.num_data_points_per_tree = self.n_points_per_tree
        self.rf = regression.binary_rss_forest()
        self.rf.options = self.rf_opts
        data = self.__init_data_container(self.X, self.y)
        self.rf.fit(data, rng=self.rng)

        return self

    def __init_data_container(self, X: np.ndarray, y: np.ndarray):
        """
        Biggest difference to SMAC3s EPM. We fit the forrest on a transformation and predict the untransformed result.
        Fills a pyrfr default data container, s.t. the forest knows
        categoricals and bounds for continous data

        Parameters
        ----------
        X : np.ndarray [n_samples, n_features]
            Input data points
        y : np.ndarray [n_samples, ]
            Corresponding target values

        Returns
        -------
        data : regression.default_data_container
            The filled data container that pyrfr can interpret
        """
        # retrieve the types and the bounds from the ConfigSpace
        data = regression.default_data_container(X.shape[1])
        if self.logged_y:
            y = y.reshape((-1, 1))
            y = np.hstack((y, np.power(10, y)))

        for i, (mn, mx) in enumerate(self.bounds):
            if np.isnan(mx):
                data.set_type_of_feature(i, mn)
            else:
                data.set_bounds_of_feature(i, mn, mx)

        for row_X, row_y in zip(X, y):
            data.add_data_point(row_X, row_y)
        return data

    def _predict(self, X: np.ndarray, cov_return_type='diagonal_cov'):
        """Predict means and variances for given X.

        Parameters
        ----------
        X : np.ndarray of shape = [n_samples,
                                   n_features (config + instance features)]

        Returns
        -------
        means : np.ndarray of shape = [n_samples, 1]
            Predictive mean
        vars : np.ndarray  of shape = [n_samples, 1]
            Predictive variance
        """
        if len(X.shape) != 2:
            raise ValueError(
                'Expected 2d array, got %dd array!' % len(X.shape))
        if X.shape[1] != len(self._initial_types):
            raise ValueError('Rows in X should have %d entries but have %d!' %
                             (len(self._initial_types), X.shape[1]))
        if cov_return_type != 'diagonal_cov':
            raise ValueError("'cov_return_type' can only take 'diagonal_cov' for this model")

        means, vars_ = [], []

        X = self._impute_inactive(X)

        for row_X in X:
            mean, var = self.rf.predict_mean_var(row_X)
            means.append(mean)
            vars_.append(var)
        means = np.array(means)
        vars_ = np.array(vars_)

        return means.reshape((-1, 1)), vars_.reshape((-1, 1))



from setuptools import setup

with open("pimp/__version__.py") as fh:
    version = fh.readlines()[-1].split()[-1].strip("\"'")

setup(
    name='PyImp',
    version=version,
    packages=['pimp', 'pimp.epm', 'pimp.utils', 'pimp.utils.io', 'pimp.evaluator', 'pimp.importance',
              'pimp.configspace'],
    entry_points={
        'console_scripts': ['pimp=pimp.pimp:cmd_line_call'],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
    ],
    platforms=['Linux'],
    install_requires=[
        "numpy",
        "sklearn",
        "matplotlib",
        "ConfigSpace>=0.4",
        "scipy",
        "pyrfr>=0.8.0",
        "smac>=0.8.0",
        "fanova",
        "tqdm",
        "argcomplete",
        "pandas",
        "bokeh>=1.1.0",
    ],
    url='',
    license='BSD 3-clause',
    author='biedenka',
    author_email='biedenka@cs.uni-freiburg.de',
    description='Package for automated Parameter Importance Analysis after Configuration.'
)



import time
from collections import OrderedDict

import matplotlib as mpl
import numpy as np
from bokeh.models import ColumnDataSource, CDSView, Span, BasicTickFormatter, Row, IndexFilter
from bokeh.plotting import figure

from pimp.utils.bokeh_helpers import shorten_unique, save_and_show

mpl.use('Agg')
from matplotlib import pyplot as plt
from tqdm import tqdm, trange
from sklearn.model_selection import KFold
from sklearn.metrics import mean_squared_error
from pimp.evaluator.base_evaluator import AbstractEvaluator

__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


class ForwardSelector(AbstractEvaluator):

    def __init__(self, scenario, cs, model, to_evaluate: int, rng, feature_imp: bool=False,
                 cv: bool=False, **kwargs):
        """
        Constructor
        :parameter:
        scenario
            SMAC scenario object
        cs
            ConfigurationSpace object
        model
            SMACs EPM (RF)
        to_evaluate
            int. Indicates for how many parameters the Importance values have to be computed
        """
        super().__init__(scenario, cs, model, to_evaluate, rng, **kwargs)
        self.name = 'Forward-Selection'
        self.logger = 'pimp.' + self.name
        self.feature_importance = feature_imp
        self.cv = cv
        self.kf = None
        self.logger.info('%snalyzing feature importance' % ('A' if feature_imp else 'Not a'))

    def _get_error(self, used, used_bounds):
        if self.cv:
            cv_errors = []
            for train_idx, test_idx, in self.kf.split(self.X):
                self._refit_model(self.types[used], self.bounds[used_bounds],
                                  self.X[train_idx.reshape(-1, 1), used],
                                  self.y[train_idx])  # refit the model every round
                pred = self.model.predict(self.X[test_idx.reshape(-1, 1), used])[0]
                cv_errors.append(np.sqrt(mean_squared_error(self.y[test_idx], pred)))
            cve = np.mean(cv_errors)
        else:
            cve = np.float('inf')
        self._refit_model(self.types[used], self.bounds[used_bounds],
                          self.X[:, used],
                          self.y)  # refit the model every round
        oob = self.model.rf.out_of_bag_error()
        return oob, cve

    def run(self) -> OrderedDict:
        """
        Implementation of the forward selection loop.
        Uses SMACs EPM (RF) wrt the configuration space to minimize the OOB error.
        :return:
        OrderedDict
            dict_keys (first key -> most important) -> OOB error
        """
        params = self.cs.get_hyperparameters()
        param_ids = list(range(len(params)))
        feature_ids = list(range(len(params), len(self.types)))
        used = []
        used_bounds = []
        if self.feature_importance:
            used.extend(range(0, len(self.bounds)))
            used_bounds.extend(range(0, len(self.bounds)))
            # names = list(map(lambda x: 'Feat #' + str(len(feature_ids) + x - len(self.types)), feature_ids))
            names = self.scenario.feature_names
            ids = feature_ids
            if self._to_eval <= 0:
                self.to_evaluate = len(feature_ids)
            else:
                self.to_evaluate = min(self._to_eval, len(ids))
        else:
            used.extend(range(len(params), len(self.model.types)))  # we don't want to evaluate the feature importance
            names = params
            ids = param_ids

        self.kf = KFold(n_splits=5)
        last_error = np.inf

        pbar = tqdm(range(self.to_evaluate), ascii=True,
                    desc='{: >.30s}: {: >7.4f} ({:s})'.format('None', np.inf, 'CV-RMSE' if self.cv else 'OOB'),
                    disable=not self.verbose)
        for round_ in pbar:  # Main Loop
            errors = []
            innerpbar = trange(len(names) + 1, ascii=True, desc='{:<40s}'.format(' '), leave=False, position=-1,
                               disable=not self.verbose)
            for idx, name in zip(ids, names):
                innerpbar.set_description('{:<40s}'.format(name if self.feature_importance else name.name))
                self.logger.debug('Evaluating %s' % name)
                used.append(idx)
                if not self.feature_importance:
                    used_bounds.append(idx)
                # self.logger.debug('Used parameters: %s' % str(used))
                # self.logger.debug('Used bounds of parameters: %s' % str(used_bounds))

                start = time.time()
                oob, cve = self._get_error(sorted(used), sorted(used_bounds))
                errors.append(cve if self.cv else oob)
                used.pop()
                if not self.feature_importance:
                    used_bounds.pop()
                self.logger.debug('Refitted RF (sec %.2f; CV-RMSE: %.4f, OOB: %.4f)' % (time.time() - start, cve, oob))
                innerpbar.update(1)
            else:  # Don't change the used/used_bounds -> add a none round
                innerpbar.set_description('{:<40s}'.format('None'))
                self.logger.debug('Evaluating None')
                start = time.time()
                oob, cve = self._get_error(sorted(used), sorted(used_bounds))
                errors.append(cve if self.cv else oob)
                self.logger.debug('Refitted RF (sec %.2f; CV-RMSE: %.4f, OOB: %.4f)' % (time.time() - start, cve, oob))
                if round_ == 0:  # Always keep track of the first None!
                    self.evaluated_parameter_importance['None'] = errors[-1]
                innerpbar.update(1)
            best_idx = np.argmin(errors)  # type: int
            lowest_error = errors[best_idx]

            if (best_idx == len(errors) - 1 or lowest_error >= last_error):
                # None was best, i.e. adding features did not improve
                pbar.set_description('{: >.30s}: {: >7.4f} ({:s})'.format('None', lowest_error,
                                                                          'CV-RMSE' if self.cv else 'OOB'))
                self.logger.info('Best result if None was added -> Early stopping')
                break
            else:
                last_error = lowest_error
                best_parameter = names.pop(best_idx)
                used.append(ids.pop(best_idx))
            if not self.feature_importance:
                used_bounds.append(used[-1])

            if self.feature_importance:
                self.logger.debug('%s: %.4f' % (best_parameter, lowest_error))
                self.evaluated_parameter_importance[best_parameter] = lowest_error
                pbar.set_description('{: >.30s}: {: >7.4f} ({:s})'.format(best_parameter, lowest_error,
                                                                          'CV-RMSE' if self.cv else 'OOB'))
            else:
                pbar.set_description('{: >.30s}: {: >7.4f} ({:s})'.format(best_parameter.name, lowest_error,
                                                                          'CV-RMSE' if self.cv else 'OOB'))
                self.evaluated_parameter_importance[best_parameter.name] = lowest_error
        all_res = {'imp': self.evaluated_parameter_importance, 'order': list(self.evaluated_parameter_importance.keys())}
        return all_res

    def _plot_result(self, name, bar=True, show=True):
        """
            plot oob score as bar charts
            Parameters
            ----------
            name
                file name to save plot
        """

        fig, ax = plt.subplots()
        params = list(self.evaluated_parameter_importance.keys())
        errors = list(self.evaluated_parameter_importance.values())
        max_to_plot = min(len(errors), self.MAX_PARAMS_TO_PLOT)

        ind = np.arange(len(errors))
        if bar:
            ax.bar(ind, errors, color=self.area_color)
        else:
            ax.plot(ind, errors, **self.LINE_FONT)

        ax.set_ylabel('CV-RMSE' if self.cv else 'OOB', **self.LABEL_FONT)
        if bar:
            ax.set_xticks(ind)
            ax.set_xlim(-.5, max_to_plot - 0.5)
        else:
            ax.set_xticks(ind)
            ax.set_xlim(0, max_to_plot - 1)
        ax.set_xticklabels(params, rotation=30, ha='right', **self.AXIS_FONT)
        ax.xaxis.grid(True)
        ax.yaxis.grid(True)

        plt.tight_layout()
        if name is not None:
            fig.savefig(name)
            if show:
                plt.show()
        else:
            plt.show()

    def plot_result(self, name=None, show=True):
        self._plot_result(name + '-barplot.png', True, show)
        self._plot_result(name + '-chng.png', False, show)
        plt.close('all')
        self.logger.info('Saved plot as %s-[barplot|chng].png' % name)

    def plot_bokeh(self, plot_name=None, show_plot=False, barplot=True, chng=True):
        """
        Plot OOB-scores as bar- or linechart in bokeh.

        Parameters
        ----------
        plot_name: str
            path where to store the plot, None to not save it
        show_plot: bool
            whether or not to open plot in standard browser
        barplot: bool
            plot OOB as barchart
        chng: bool
            plot OOB as linechart

        Returns
        -------
        layout: bokeh.models.Row
            bokeh plot (can be used in notebook or comparted with components)
        """
        # Get all relevant data-points
        params = list(self.evaluated_parameter_importance.keys())
        p_names_short = shorten_unique(params)
        errors = list(self.evaluated_parameter_importance.values())
        max_to_plot = min(len(errors), self.MAX_PARAMS_TO_PLOT)
        plot_indices = sorted(range(len(errors)), key=lambda x: errors[x], reverse=True)[:max_to_plot]

        # Customizing plot-style
        bar_width = 25

        # Create ColumnDataSource for both plots
        source = ColumnDataSource(data=dict(parameter_names_short=p_names_short,
                                            parameter_names=params,
                                            parameter_importance=errors,
                                            ))
        plots = []

        view = CDSView(source=source, filters=[IndexFilter(plot_indices)])
        tooltips = [("Parameter", "@parameter_names"),
                    ("Importance", "@parameter_importance"),
                    ]
        if barplot:
            p = figure(x_range=p_names_short,
                       plot_height=350, plot_width=100 + max_to_plot * bar_width,
                       toolbar_location=None, tools="hover", tooltips=tooltips)

            p.vbar(x='parameter_names_short', top='parameter_importance', width=0.9, source=source, view=view)
            for value in [self.IMPORTANCE_THRESHOLD, -self.IMPORTANCE_THRESHOLD]:
                p.add_layout(Span(location=value, dimension='width', line_color='red', line_dash='dashed'))

            plots.append(p)

        if chng:
            p = figure(x_range=p_names_short,
                       plot_height=350, plot_width=100 + max_to_plot * bar_width,
                       toolbar_location=None, tools="hover", tooltips=tooltips)

            p.line(x='parameter_names_short', y='parameter_importance', source=source, view=view)

            plots.append(p)

        # Common styling:
        for p in plots:
            p.xaxis.major_label_orientation = 1.3
            p.xaxis.major_label_text_font_size = "14pt"
            p.yaxis.formatter = BasicTickFormatter(use_scientific=False)
            p.yaxis.axis_label = 'CV-RMSE' if self.cv else 'OOB'

        layout = Row(*plots)

        # Save and show...
        save_and_show(plot_name, show_plot, layout)

        return layout



# PYTHON_ARGCOMPLETE_OK
import os
if not 'MATPLOTLIBRC' in os.environ:
    os.environ['MATPLOTLIBRC'] = os.path.join(os.path.dirname(__file__))
import json
import datetime
import logging
import time
import warnings
from typing import List
from typing import Dict
from typing import Union

import numpy as np
from smac.scenario.scenario import Scenario
from smac.runhistory.runhistory import RunHistory, RunKey
from smac.tae.execute_ta_run_aclib import StatusType
from ConfigSpace.configuration_space import Configuration
from pimp.importance.importance import Importance
from pimp.utils.io.cmd_reader import CMDs

from smac.facade.smac_ac_facade import SMAC4AC as SMAC

__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


class PIMP:
    def __init__(self,
                 scenario: Scenario,
                 smac: Union[SMAC, None] = None,
                 mode: str = 'all',
                 X: Union[None, List[list], np.ndarray] = None,
                 y: Union[None, List[list], np.ndarray] = None,
                 numParams: int = -1,
                 impute: bool = False,
                 seed: int = 12345,
                 run: bool = False,
                 max_sample_size: int = -1,
                 fanova_cut_at_default: bool = False,
                 fANOVA_pairwise: bool = True,
                 forwardsel_feat_imp: bool = False,
                 incn_quant_var: bool = True,
                 marginalize_away_instances: bool = False,
                 save_folder: str = 'PIMP'):
        """
        Interface to be used with SMAC or with X and y matrices.
        :param scenario: The scenario object, that knows the configuration space.
        :param smac: The smac object that keeps all the run-data
        :param mode: The mode with which to run PIMP [ablation, fanova, all, forward-selection]
        :param X: Numpy Array that contains parameter arrays
        :param y: Numpy array that contains the corresponding performance values
        :param numParams: The number of parameters to evaluate
        :param impute: Flag to decide if censored data gets imputed or not
        :param seed: The random seed
        :param run: Flag to immediately compute the importance values after this setup or not.
        """
        self.scenario = scenario
        self.imp = None
        self.mode = mode
        self.save_folder = save_folder
        if not os.path.exists(self.save_folder): os.mkdir(self.save_folder)
        if smac is not None:
            self.imp = Importance(scenario=scenario,
                                  runhistory=smac.runhistory,
                                  incumbent=smac.solver.incumbent,
                                  seed=seed,
                                  parameters_to_evaluate=numParams,
                                  save_folder='PIMP',
                                  impute_censored=impute,
                                  max_sample_size=max_sample_size,
                                  fANOVA_cut_at_default=fanova_cut_at_default,
                                  fANOVA_pairwise=fANOVA_pairwise,
                                  forwardsel_feat_imp=forwardsel_feat_imp,
                                  incn_quant_var=incn_quant_var,
                                  preprocess=marginalize_away_instances)
        elif X is not None and y is not None:
            X = np.array(X)
            y = np.array(y)
            runHist = RunHistory()
            if X.shape[0] != y.shape[0]:
                raise Exception('Number of samples in X and y dont match!')
            n_params = len(scenario.cs.get_hyperparameters())
            feats = None
            if X.shape[1] > n_params:
                feats = X[:, n_params:]
                assert feats.shape[1] == scenario.feature_array.shape[1]
                X = X[:, :n_params]

            for p in range(X.shape[1]):  # Normalize the data to fit into [0, 1]
                _min, _max = np.min(X[:, p]), np.max(X[:, p])
                if _min < 0. or 1 < _max:  # if it is not already normalized
                    for id, v in enumerate(X[:, p]):
                        X[id, p] = (v - _min) / (_max - _min)

            # Add everything to a runhistory such that PIMP can work with it
            for x, feat, y_val in zip(X, feats if feats is not None else X, y):
                id = None
                for inst in scenario.feature_dict:  # determine on which instance a configuration was run
                    if np.all(scenario.feature_dict[inst] == feat):
                        id = inst
                        break
                runHist.add(Configuration(scenario.cs, vector=x), y_val, 0, StatusType.SUCCESS, id)
            self.X = X
            self.y = y

            best_ = None  # Determine incumbent according to the best mean cost in the runhistory
            for config in runHist.config_ids:
                inst_seed_pairs = runHist.get_runs_for_config(config)
                all_ = []
                for inst, seed in inst_seed_pairs:
                    rk = RunKey(runHist.config_ids[config], inst, seed)
                    all_.append(runHist.data[rk].cost)
                mean = np.mean(all_)
                if best_ is None or best_[0] > mean:
                    best_ = (mean, config)
            incumbent = best_[1]
            self.imp = Importance(scenario=scenario,
                                  runhistory=runHist,
                                  seed=seed,
                                  parameters_to_evaluate=numParams,
                                  save_folder=self.save_folder,
                                  impute_censored=impute,
                                  incumbent=incumbent,
                                  fANOVA_cut_at_default=fanova_cut_at_default,
                                  fANOVA_pairwise=fANOVA_pairwise,
                                  forwardsel_feat_imp=forwardsel_feat_imp,
                                  incn_quant_var=incn_quant_var,
                                  preprocess=marginalize_away_instances
                                  )
        else:
            raise Exception('Neither X and y matrices nor a SMAC object were specified to compute the importance '
                            'values from!')

        if run:
            self.compute_importances()

    def compute_importances(self):
        if self.mode == 'all':
            self.mode = ['ablation',
                         'forward-selection',
                         'fanova',
                         'incneighbor']
        elif not isinstance(self.mode, list):
            self.mode = [self.mode]
        result = self.imp.evaluate_scenario(self.mode, save_folder=self.save_folder)
        return result

    def plot_results(self, result: Union[List[Dict[str, float]], Dict[str, float]], save_table: bool = True,
                     show=False):
        save_folder = self.save_folder
        if self.mode == 'all':
            with open(os.path.join(save_folder, 'pimp_values_%s.json' % self.mode), 'w') as out_file:
                json.dump(result[0], out_file, sort_keys=True, indent=4, separators=(',', ': '))
            self.imp.plot_results(list(map(lambda x: os.path.join(save_folder, x.name.lower()), result[1])),
                                  result[1], show=show)
            if save_table:
                self.imp.table_for_comparison(evaluators=result[1], name=os.path.join(
                    save_folder, 'pimp_table_%s.tex' % self.mode), style='latex')
            else:
                self.imp.table_for_comparison(evaluators=result[1], style='cmd')
        else:
            with open(os.path.join(save_folder, 'pimp_values_%s.json' % self.mode), 'w') as out_file:
                json.dump(result[0], out_file, sort_keys=True, indent=4, separators=(',', ': '))
            if isinstance(self.mode, list):
                self.imp.plot_results(name=os.path.join(save_folder, 'all'), show=show)
            else:
                self.imp.plot_results(name=os.path.join(save_folder, self.mode), show=show)


def cmd_line_call():
    """
    Main Parameter importance script.
    """
    cmd_reader = CMDs()
    args, misc_ = cmd_reader.read_cmd()  # read cmd args
    cwd = os.path.abspath(os.getcwd())
    if args.out_folder and not os.path.isabs(args.out_folder):
        args.out_folder = os.path.abspath(args.out_folder)
    if args.trajectory and not os.path.isabs(args.trajectory):
        args.trajectory = os.path.abspath(args.trajectory)
    if not os.path.isabs(args.scenario_file):
        args.scenario_file = os.path.abspath(args.scenario_file)
    if not os.path.isabs(args.history):
        args.history = os.path.abspath(args.history)
    os.chdir(args.wdir)
    logging.basicConfig(level=args.verbose_level)
    ts = time.time()
    ts = datetime.datetime.fromtimestamp(ts).strftime('%Y_%m_%d_%H:%M:%S')
    fanova_ready = True

    try:
        import fanova
    except ImportError:
        warnings.simplefilter('always', ImportWarning)
        warnings.warn('fANOVA is not installed in your environment. To install it please run '
                      '"git+http://github.com/automl/fanova.git@master"')
        fanova_ready = False

    if 'influence-model' in args.modus:
        logging.warning('influence-model not fully supported yet!')
    if 'incneighbor' in args.modus:
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn('incneighbor will be deprecated in version 1.0.0 as it was the development name of'
                      ' lpi. Use lpi instead.', DeprecationWarning, stacklevel=2)
    if 'lpi' in args.modus:  # LPI will replace incneighbor in the future
        args.modus[args.modus.index('lpi')] = 'incneighbor'
    if 'fanova' in args.modus and not fanova_ready:
        raise ImportError('fANOVA is not installed! To install it please run '
                          '"git+http://github.com/automl/fanova.git@master"')
    if 'all' in args.modus:
        choices = ['ablation',
                   'forward-selection',
                   'fanova',
                   'incneighbor']
        if not fanova_ready:
            raise ImportError('fANOVA is not installed! To install it please run '
                              '"git+http://github.com/automl/fanova.git@master"')
        del args.modus[args.modus.index('all')]
        if len(args.modus) == len(choices):
            pass
        else:
            args.modus = choices
    if not args.out_folder:
        if len(args.modus) > 1:
            tmp = ['all']
        else:
            tmp = args.modus
            if 'incneighbor' in args.modus:
                tmp = ['lpi']
        save_folder = os.path.join(cwd, 'PIMP_%s' % '_'.join(tmp))
        if os.path.exists(os.path.abspath(save_folder)):
            save_folder = os.path.join(cwd, 'PIMP_%s_%s' % ('_'.join(tmp), ts))
    else:
        if len(args.modus) > 1:
            tmp = ['all']
        else:
            tmp = args.modus
            if 'incneighbor' in args.modus:
                tmp = ['lpi']
        if os.path.exists(os.path.abspath(args.out_folder)) or os.path.exists(os.path.abspath(
                        args.out_folder + '_%s' % '_'.join(tmp))):
            save_folder = os.path.join(cwd, args.out_folder + '_%s_%s' % ('_'.join(tmp), ts))
        else:
            save_folder = os.path.join(cwd, args.out_folder + '_%s' % '_'.join(tmp))

    importance = Importance(scenario_file=args.scenario_file,
                            runhistory_file=args.history,
                            parameters_to_evaluate=args.num_params,
                            traj_file=args.trajectory, seed=args.seed,
                            save_folder=save_folder,
                            impute_censored=args.impute,
                            max_sample_size=args.max_sample_size,
                            fANOVA_cut_at_default=args.fanova_cut_at_default,
                            fANOVA_pairwise=args.fanova_pairwise,
                            forwardsel_feat_imp=args.forwardsel_feat_imp,
                            incn_quant_var=args.incn_quant_var,
                            preprocess=args.marg_inst,
                            forwardsel_cv=args.forwardsel_cv)  # create importance object
    with open(os.path.join(save_folder, 'pimp_args.json'), 'w') as out_file:
        json.dump(args.__dict__, out_file, sort_keys=True, indent=4, separators=(',', ': '))
    result = importance.evaluate_scenario(args.modus, save_folder=save_folder)
    if args.table:
        importance.table_for_comparison(evaluators=result[1], name=os.path.join(
            save_folder, 'pimp_table_%s.tex' % args.modus), style='latex')
    else:
        importance.table_for_comparison(evaluators=result[1], style='cmd')
    os.chdir(cwd)


if __name__ == '__main__':
    cmd_line_call()



import numpy as np
from scipy import stats

from pimp.epm.base_epm import RandomForestWithInstances as rfi

__author__ = "Andre Biedenkapp"
__copyright__ = "Copyright 2016, ML4AAD"
__license__ = "3-clause BSD"
__maintainer__ = "Andre Biedenkapp"
__email__ = "biedenka@cs.uni-freiburg.de"


class EPARrfi(rfi):

    def __init__(self, configspace, types, bounds, seed,
                 cutoff=0,
                 threshold=0, **kwargs):
        """
        Interface to the random forest that takes instance features
        into account.

        Parameters
        ----------
        configspace: ConfigurationSpace
            configspace to be passed to random forest (used to impute inactive parameter-values)
        types: np.ndarray (D)
            Specifies the number of categorical values of an input dimension. Where
            the i-th entry corresponds to the i-th input dimension. Let say we have
            2 dimension where the first dimension consists of 3 different
            categorical choices and the second dimension is continuous than we
            have to pass np.array([2, 0]). Note that we count starting from 0.
        bounds: np.ndarray (D)
            Specifies the bounds
        seed: int
            The seed that is passed to the random_forest_run library.
        instance_features: np.ndarray (I, K)
            Contains the K dimensional instance features
            of the I different instances
        num_trees: int
            The number of trees in the random forest.
        do_bootstrapping: bool
            Turns on / off bootstrapping in the random forest.
        ratio_features: float
            The ratio of features that are considered for splitting.
        min_samples_split: int
            The minimum number of data points to perform a split.
        min_samples_leaf: int
            The minimum number of data points in a leaf.
        max_depth: int

        eps_purity: float

        max_num_nodes: int

        cutoff: int
            The cutoff used in the specified scenario

        threshold:
            Maximal possible value
        """
        super().__init__(configspace=configspace, types=types, bounds=bounds, seed=seed, **kwargs)
        np.seterr(divide='ignore', invalid='ignore')
        self.cutoff = cutoff
        self.threshold = threshold

    def _predict_EPAR(self, mean_var, prediction_threshold=0):
        """
        Predicting the Expected Penalized Average Runtime according to the cuttoff and par-factor specified in the
        scenario.

        Parameters
        ----------
        X : np.ndarray of shape = [n_samples, n_features (config + instance
        features)]
        """

        pred = np.zeros(shape=mean_var[0].shape)
        var = np.zeros(shape=mean_var[1].shape)

        for p in range(pred.shape[0]):
            var[p] = mean_var[1][p]
            if mean_var[0][p] > self.cutoff:
                # mean prediction is already higher than cutoff
                self.logger.debug("Predicted %g which is higher than cutoff"
                                  " %s" % (mean_var[0][p], self.cutoff))
                # pred[p] = tmp_threshold
                # continue

            # Calc cdf from -inf to cutoff
            cdf = stats.norm.cdf(x=self.cutoff, loc=mean_var[0][p],
                                 scale=np.sqrt(mean_var[1][p]))

            # Probability mass > cutoff
            upper_exp = 1 - cdf

            if upper_exp > 1:
                self.logger.warn("Upper exp is larger than 1, "
                                 "is this possible: %g > 1" % upper_exp)
                upper_exp = 1
                cdf = 0

            if upper_exp < prediction_threshold or self.threshold == self.cutoff:
                # There is not enough probability mass higher than cutoff
                # Or threshold == cutoff
                pred[p] = mean_var[0][p]
            else:
                # Calculate mean of lower truncnorm
                lower_pred = stats.truncnorm.stats(
                    a=(-np.inf - mean_var[0][p]) / np.sqrt(mean_var[1][p]),
                    b=(self.cutoff - mean_var[0][p]) / np.sqrt(mean_var[1][p]),
                    loc=mean_var[0][p],
                    scale=np.sqrt(mean_var[1][p]),
                    moments='m')

                upper_pred = upper_exp * self.threshold
                pred[p] = lower_pred * cdf + upper_pred

                if pred[p] > self.threshold + 10 ** -5:
                    raise ValueError("Predicted higher than possible, %g > %g"
                                     % (pred[p], self.threshold))

                # This can happen and if it happens, set prediction to cutoff
                if not np.isfinite(pred[p]):
                    self.logger.debug("Prediction is not finite cdf %g, "
                                      "lower_pred %g; Setting %g to %g" %
                                      (cdf, lower_pred, pred[p],
                                       self.cutoff + 10 ** -5))
                    pred[p] = self.cutoff + 10 ** -5
        return pred, var

    def predict(self, X):
        """
        Method to override the predict method of RandomForestWithInstances.
        Thus it can be used in the marginalized over instances method of the RFWI class
        """
        return self._predict_EPAR(self._predict(X))
